@@ -88,6 +88,7 @@ struct _GcrMockPromptClass {
 };
 
 typedef struct {
+	gboolean close;
 	gboolean proceed;
 	gchar *password;
 	GList *properties;
@@ -104,7 +105,7 @@ typedef struct {
 
 	/* Owned by the prompter thread*/
 	GcrSystemPrompter *prompter;
-	const gchar *bus_name;
+	GDBusConnection *connection;
 	GMainLoop *loop;
 } ThreadData;
 
@@ -396,10 +397,21 @@ _gcr_mock_prompt_class_init (GcrMockPromptClass *klass)
 }
 
 static gboolean
-on_timeout_complete_response (gpointer data)
+on_timeout_complete (gpointer data)
 {
 	GSimpleAsyncResult *res = data;
-	g_simple_async_result_complete_in_idle (res);
+	g_simple_async_result_complete (res);
+	return FALSE;
+}
+
+static gboolean
+on_timeout_complete_and_close (gpointer data)
+{
+	GSimpleAsyncResult *res = data;
+	GcrPrompt *prompt = GCR_PROMPT (g_async_result_get_source_object (data));
+	g_simple_async_result_complete (res);
+	gcr_prompt_close (prompt);
+	g_object_unref (prompt);
 	return FALSE;
 }
 
@@ -410,8 +422,10 @@ gcr_mock_prompt_confirm_async (GcrPrompt *prompt,
                                gpointer user_data)
 {
 	GcrMockPrompt *self = GCR_MOCK_PROMPT (prompt);
+	GSourceFunc complete_func = on_timeout_complete;
 	GSimpleAsyncResult *res;
 	MockResponse *response;
+	GSource *source;
 	guint delay_msec;
 
 	g_mutex_lock (running->mutex);
@@ -426,6 +440,10 @@ gcr_mock_prompt_confirm_async (GcrPrompt *prompt,
 		g_critical ("password prompt requested, but not expected");
 		g_simple_async_result_set_op_res_gboolean (res, FALSE);
 
+	} else if (response->close) {
+		complete_func = on_timeout_complete_and_close;
+		g_simple_async_result_set_op_res_gboolean (res, FALSE);
+
 	} else if (response->password) {
 		g_critical ("confirmation prompt requested, but password prompt expected");
 		g_simple_async_result_set_op_res_gboolean (res, FALSE);
@@ -435,17 +453,17 @@ gcr_mock_prompt_confirm_async (GcrPrompt *prompt,
 		g_simple_async_result_set_op_res_gboolean (res, response->proceed);
 	}
 
+	if (delay_msec > 0)
+		source = g_timeout_source_new (delay_msec);
+	else
+		source = g_idle_source_new ();
+
+	g_assert (!self->delay_source);
+	g_source_set_callback (source, complete_func, g_object_ref (res), g_object_unref);
+	self->delay_source = g_source_attach (source, g_main_context_get_thread_default ());
+	g_source_unref (source);
+
 	mock_response_free (response);
-
-	if (delay_msec > 0) {
-		g_assert (!self->delay_source);
-		self->delay_source = g_timeout_add_full (G_PRIORITY_DEFAULT, delay_msec,
-		                                         on_timeout_complete_response,
-		                                         g_object_ref (res), g_object_unref);
-	} else {
-		on_timeout_complete_response (res);
-	}
-
 	g_object_unref (res);
 }
 
@@ -484,8 +502,10 @@ gcr_mock_prompt_password_async (GcrPrompt *prompt,
                                 gpointer user_data)
 {
 	GcrMockPrompt *self = GCR_MOCK_PROMPT (prompt);
+	GSourceFunc complete_func = on_timeout_complete;
 	GSimpleAsyncResult *res;
 	MockResponse *response;
+	GSource *source;
 	guint delay_msec;
 
 	g_mutex_lock (running->mutex);
@@ -500,6 +520,10 @@ gcr_mock_prompt_password_async (GcrPrompt *prompt,
 		g_critical ("password prompt requested, but not expected");
 		g_simple_async_result_set_op_res_gpointer (res, NULL, NULL);
 
+	} else if (response->close) {
+		g_simple_async_result_set_op_res_gpointer (res, NULL, NULL);
+		complete_func = on_timeout_complete_and_close;
+
 	} else if (!response->password) {
 		g_critical ("password prompt requested, but confirmation prompt expected");
 		g_simple_async_result_set_op_res_gpointer (res, NULL, NULL);
@@ -517,14 +541,15 @@ gcr_mock_prompt_password_async (GcrPrompt *prompt,
 
 	mock_response_free (response);
 
-	if (delay_msec > 0) {
-		g_assert (!self->delay_source);
-		self->delay_source = g_timeout_add_full (G_PRIORITY_DEFAULT, delay_msec,
-		                                         on_timeout_complete_response,
-		                                         g_object_ref (res), g_object_unref);
-	} else {
-		on_timeout_complete_response (res);
-	}
+	if (delay_msec > 0)
+		source = g_timeout_source_new (delay_msec);
+	else
+		source = g_idle_source_new ();
+
+	g_assert (!self->delay_source);
+	g_source_set_callback (source, complete_func, g_object_ref (res), g_object_unref);
+	self->delay_source = g_source_attach (source, g_main_context_get_thread_default ());
+	g_source_unref (source);
 
 	g_object_unref (res);
 }
@@ -792,6 +817,30 @@ gcr_mock_prompter_expect_password_cancel (void)
 	g_mutex_unlock (running->mutex);
 }
 
+/**
+ * gcr_mock_prompter_expect_close:
+ *
+ * Queue an expected response on the mock prompter.
+ *
+ * Expects any prompt, and closes the prompt when it gets it.
+ */
+void
+gcr_mock_prompter_expect_close (void)
+{
+	MockResponse *response;
+
+	g_assert (running != NULL);
+
+	g_mutex_lock (running->mutex);
+
+	response = g_new0 (MockResponse, 1);
+	response->close = TRUE;
+
+	g_queue_push_tail (&running->responses, response);
+
+	g_mutex_unlock (running->mutex);
+}
+
 /**
  * gcr_mock_prompter_is_expecting:
  *
@@ -894,9 +943,9 @@ mock_prompter_thread (gpointer data)
 		                                                     G_DBUS_CONNECTION_FLAGS_MESSAGE_BUS_CONNECTION,
 		                                                     NULL, NULL, &error);
 		if (error == NULL) {
+			thread_data->connection = connection;
 			gcr_system_prompter_register (GCR_SYSTEM_PROMPTER (thread_data->prompter),
 			                              connection);
-			thread_data->bus_name = g_dbus_connection_get_unique_name (connection);
 		} else {
 			g_critical ("couldn't create connection: %s", error->message);
 			g_error_free (error);
@@ -929,10 +978,19 @@ mock_prompter_thread (gpointer data)
 	thread_data->prompter = NULL;
 
 	if (connection) {
-		if (!g_dbus_connection_flush_sync (connection, NULL, &error)) {
-			g_critical ("connection flush failed: %s", error->message);
-			g_error_free (error);
+		thread_data->connection = NULL;
+
+		if (!g_dbus_connection_is_closed (connection)) {
+			if (!g_dbus_connection_flush_sync (connection, NULL, &error)) {
+				g_critical ("connection flush failed: %s", error->message);
+				g_error_free (error);
+			}
+			if (!g_dbus_connection_close_sync (connection, NULL, &error)) {
+				g_critical ("connection close failed: %s", error->message);
+				g_error_free (error);
+			}
 		}
+
 		g_object_unref (connection);
 	}
 
@@ -981,7 +1039,22 @@ gcr_mock_prompter_start (void)
 	g_assert (running->prompter);
 	g_mutex_unlock (running->mutex);
 
-	return running->bus_name;
+	return g_dbus_connection_get_unique_name (running->connection);
+}
+
+void
+gcr_mock_prompter_disconnect (void)
+{
+	GError *error = NULL;
+
+	g_assert (running != NULL);
+	g_assert (running->connection);
+
+	g_dbus_connection_close_sync (running->connection, NULL, &error);
+	if (error != NULL) {
+		g_critical ("disconnect connection close failed: %s", error->message);
+		g_error_free (error);
+	}
 }
 
 /**
