@@ -143,9 +143,10 @@ static gint unique_prompt_id = 0;
 
 typedef struct {
 	GSource *timeout;
+	GSource *waiting;
 	GMainContext *context;
 	GCancellable *cancellable;
-	gboolean waiting;
+	guint watch_id;
 } CallClosure;
 
 static void
@@ -154,10 +155,44 @@ call_closure_free (gpointer data)
 	CallClosure *closure = data;
 	if (closure->timeout)
 		g_source_destroy (closure->timeout);
-	g_clear_object (&closure->cancellable);
+	if (closure->waiting)
+		g_source_destroy (closure->waiting);
+	if (closure->watch_id)
+		g_bus_unwatch_name (closure->watch_id);
+	g_object_unref (closure->cancellable);
 	g_free (data);
 }
 
+static void
+on_propagate_cancelled (GCancellable *cancellable,
+                        gpointer user_data)
+{
+	/* Propagate the cancelled signal */
+	GCancellable *cancel = G_CANCELLABLE (user_data);
+	g_cancellable_cancel (cancel);
+}
+
+static CallClosure *
+call_closure_new (GCancellable *cancellable)
+{
+	CallClosure *call;
+
+	/*
+	 * We use our own cancellable object, since we cancel it it in
+	 * situations other than when the caller cancels.
+	 */
+
+	call = g_new0 (CallClosure, 1);
+	call->cancellable = g_cancellable_new ();
+
+	if (cancellable) {
+		g_cancellable_connect (cancellable, G_CALLBACK (on_propagate_cancelled),
+		                       g_object_ref (call->cancellable), g_object_unref);
+	}
+
+	return call;
+}
+
 static void
 gcr_system_prompt_init (GcrSystemPrompt *self)
 {
@@ -390,16 +425,101 @@ gcr_system_prompt_constructed (GObject *obj)
 		self->pv->prompter_bus_name = g_strdup (GCR_DBUS_PROMPTER_SYSTEM_BUS_NAME);
 }
 
+static void
+on_prompter_stop_prompting (GObject *source,
+                            GAsyncResult *result,
+                            gpointer user_data)
+{
+	GSimpleAsyncResult *async = NULL;
+	GError *error = NULL;
+	GVariant *retval;
+
+	retval = g_dbus_connection_call_finish (G_DBUS_CONNECTION (source), result, &error);
+	if (error != NULL) {
+		_gcr_debug ("failed to stop prompting: %s", egg_error_message (error));
+		g_clear_error (&error);
+	}
+
+	if (retval)
+		g_variant_unref (retval);
+
+	if (user_data) {
+		async = G_SIMPLE_ASYNC_RESULT (user_data);
+		g_simple_async_result_complete (async);
+		g_object_unref (async);
+	}
+}
+
+static void
+perform_close (GcrSystemPrompt *self,
+               GSimpleAsyncResult *async,
+               GCancellable *cancellable)
+{
+	GSimpleAsyncResult *res;
+	CallClosure *closure;
+	gboolean called = FALSE;
+	gboolean closed;
+
+	closed = self->pv->closed;
+	self->pv->closed = TRUE;
+
+	if (!closed)
+		_gcr_debug ("closing prompt");
+
+	if (self->pv->pending) {
+		res = g_object_ref (self->pv->pending);
+		g_clear_object (&self->pv->pending);
+		closure = g_simple_async_result_get_op_res_gpointer (res);
+		g_cancellable_cancel (closure->cancellable);
+		g_simple_async_result_complete_in_idle (res);
+		g_object_unref (res);
+	}
+
+	if (self->pv->prompt_registered) {
+		g_dbus_connection_unregister_object (self->pv->connection,
+		                                     self->pv->prompt_registered);
+		self->pv->prompt_registered = 0;
+	}
+
+	if (self->pv->begun_prompting) {
+		if (self->pv->connection && self->pv->prompt_path) {
+			_gcr_debug ("Calling the prompter %s method", GCR_DBUS_PROMPTER_METHOD_STOP);
+			g_dbus_connection_call (self->pv->connection,
+			                        self->pv->prompter_bus_name,
+			                        GCR_DBUS_PROMPTER_OBJECT_PATH,
+			                        GCR_DBUS_PROMPTER_INTERFACE,
+			                        GCR_DBUS_PROMPTER_METHOD_STOP,
+			                        g_variant_new ("(o)", self->pv->prompt_path),
+			                        G_VARIANT_TYPE ("()"),
+			                        G_DBUS_CALL_FLAGS_NO_AUTO_START,
+			                        -1, cancellable,
+			                        on_prompter_stop_prompting,
+			                        async ? g_object_ref (async) : NULL);
+			called = TRUE;
+		}
+		self->pv->begun_prompting = FALSE;
+	}
+
+	g_free (self->pv->prompt_path);
+	self->pv->prompt_path = NULL;
+
+	g_clear_object (&self->pv->connection);
+
+	if (!called && async)
+		g_simple_async_result_complete_in_idle (async);
+
+	/* Emit the signal if necessary, after closed */
+	if (!closed)
+		gcr_prompt_close (GCR_PROMPT (self));
+}
+
 static void
 gcr_system_prompt_dispose (GObject *obj)
 {
 	GcrSystemPrompt *self = GCR_SYSTEM_PROMPT (obj);
 
 	g_clear_object (&self->pv->exchange);
-
-	_gcr_debug ("closing prompt asynchronously: %s", self->pv->prompt_path);
-	if (self->pv->connection)
-		gcr_system_prompt_close_async (self, NULL, NULL, NULL);
+	perform_close (self, NULL, NULL);
 
 	g_hash_table_remove_all (self->pv->properties);
 	g_hash_table_remove_all (self->pv->dirty_properties);
@@ -560,17 +680,15 @@ prompt_method_done (GcrSystemPrompt *self,
                     GDBusMethodInvocation *invocation,
                     GVariant *parameters)
 {
-	GSimpleAsyncResult *res;
-
-	g_return_if_fail (G_IS_SIMPLE_ASYNC_RESULT (self->pv->pending));
-
 	g_dbus_method_invocation_return_value (invocation, g_variant_new ("()"));
 
-	res = g_object_ref (self->pv->pending);
-	g_clear_object (&self->pv->pending);
-	g_simple_async_result_set_op_res_gpointer (res, NULL, NULL);
-	g_simple_async_result_complete (res);
-	g_object_unref (res);
+	/*
+	 * At this point we're done prompting, and calling StopPrompting
+	 * on the prompter is no longer necessary. It may have already been
+	 * called, or the prompter may have stopped on its own accord.
+	 */
+	self->pv->begun_prompting = FALSE;
+	perform_close (self, NULL, NULL);
 }
 
 static void
@@ -659,6 +777,16 @@ register_prompt_object (GcrSystemPrompt *self,
 	}
 }
 
+static void
+on_prompter_vanished (GDBusConnection *connection,
+                      const gchar *name,
+                      gpointer user_data)
+{
+	GSimpleAsyncResult *async = G_SIMPLE_ASYNC_RESULT (user_data);
+	CallClosure *call = g_simple_async_result_get_op_res_gpointer (async);
+	g_cancellable_cancel (call->cancellable);
+}
+
 static void
 on_bus_connected (GObject *source,
                   GAsyncResult *result,
@@ -681,6 +809,12 @@ on_bus_connected (GObject *source,
 
 		g_main_context_push_thread_default (closure->context);
 
+		closure->watch_id = g_bus_watch_name_on_connection (self->pv->connection,
+		                                                    self->pv->prompter_bus_name,
+		                                                    G_BUS_NAME_WATCHER_FLAGS_NONE,
+		                                                    NULL, on_prompter_vanished,
+		                                                    res, NULL);
+
 		register_prompt_object (self, &error);
 
 		g_main_context_pop_thread_default (closure->context);
@@ -753,6 +887,27 @@ on_call_timeout (gpointer user_data)
 	return FALSE; /* Don't call this function again */
 }
 
+static gboolean
+on_call_cancelled (GCancellable *cancellable,
+                   gpointer user_data)
+{
+	GSimpleAsyncResult *async = G_SIMPLE_ASYNC_RESULT (user_data);
+	CallClosure *call = g_simple_async_result_get_op_res_gpointer (async);
+	GcrSystemPrompt *self = GCR_SYSTEM_PROMPT (g_async_result_get_source_object (user_data));
+
+	g_source_destroy (call->waiting);
+	call->waiting = NULL;
+
+	g_simple_async_result_set_error (async, G_IO_ERROR, G_IO_ERROR_CANCELLED,
+	                                 _("The operation was cancelled"));
+
+	/* Tell the prompter we're no longer interested */
+	gcr_system_prompt_close_async (self, NULL, NULL, NULL);
+
+	g_object_unref (self);
+	return FALSE; /* Don't call this function again */
+}
+
 void
 perform_init_async (GcrSystemPrompt *self,
                     GSimpleAsyncResult *res)
@@ -814,7 +969,7 @@ gcr_system_prompt_real_init_async (GAsyncInitable *initable,
 
 	res = g_simple_async_result_new (G_OBJECT (self), callback, user_data,
 	                                 gcr_system_prompt_real_init_async);
-	closure = g_new0 (CallClosure, 1);
+	closure = call_closure_new (cancellable);
 	closure->context = g_main_context_get_thread_default ();
 	if (closure->context)
 		g_main_context_ref (closure->context);
@@ -947,16 +1102,20 @@ on_perform_prompt_complete (GObject *source,
 {
 	GSimpleAsyncResult *res = G_SIMPLE_ASYNC_RESULT (user_data);
 	GcrSystemPrompt *self = GCR_SYSTEM_PROMPT (g_async_result_get_source_object (user_data));
-	CallClosure *closure = g_simple_async_result_get_op_res_gpointer (res);
+	CallClosure *call = g_simple_async_result_get_op_res_gpointer (res);
 	GError *error = NULL;
 	GVariant *retval;
 
 	retval = g_dbus_connection_call_finish (G_DBUS_CONNECTION (source), result, &error);
 	if (error != NULL) {
+		self->pv->pending = NULL;
 		g_simple_async_result_take_error (res, error);
-		g_simple_async_result_complete_in_idle (res);
+		g_simple_async_result_complete (res);
 	} else {
-		closure->waiting = TRUE;
+		g_assert (call->waiting == NULL);
+		call->waiting = g_cancellable_source_new (call->cancellable);
+		g_source_set_callback (call->waiting, (GSourceFunc)on_call_cancelled, res, NULL);
+		g_source_attach (call->waiting, call->context);
 	}
 
 	if (retval)
@@ -988,19 +1147,31 @@ perform_prompt_async (GcrSystemPrompt *self,
 		return;
 	}
 
-	_gcr_debug ("prompting for password");
-
 	res = g_simple_async_result_new (G_OBJECT (self), callback, user_data, source_tag);
-	closure = g_new0 (CallClosure, 1);
-	closure->cancellable = cancellable ? g_object_ref (cancellable) : cancellable;
+	closure = call_closure_new (cancellable);
 	g_simple_async_result_set_op_res_gpointer (res, closure, call_closure_free);
 
+	if (self->pv->closed) {
+		self->pv->last_response = g_strdup (GCR_DBUS_PROMPT_REPLY_NONE);
+		g_simple_async_result_complete_in_idle (res);
+		g_object_unref (res);
+		return;
+	}
+
+	_gcr_debug ("prompting for password");
+
 	exchange = gcr_system_prompt_get_secret_exchange (self);
 	if (self->pv->received)
 		sent = gcr_secret_exchange_send (exchange, NULL, 0);
 	else
 		sent = gcr_secret_exchange_begin (exchange);
 
+	closure->watch_id = g_bus_watch_name_on_connection (self->pv->connection,
+	                                                    self->pv->prompter_bus_name,
+	                                                    G_BUS_NAME_WATCHER_FLAGS_NONE,
+	                                                    NULL, on_prompter_vanished,
+	                                                    res, NULL);
+
 	builder = build_dirty_properties (self);
 
 	/* Reregister the prompt object in the current GMainContext */
@@ -1110,6 +1281,22 @@ gcr_system_prompt_confirm_finish (GcrPrompt *prompt,
 	return handle_last_response (self);
 }
 
+static void
+gcr_system_prompt_real_close (GcrPrompt *prompt)
+{
+	GcrSystemPrompt *self = GCR_SYSTEM_PROMPT (prompt);
+
+	/*
+	 * Setting this before calling close_async allows us to prevent firing
+	 * this signal again in a loop.
+	 */
+
+	if (!self->pv->closed) {
+		self->pv->closed = TRUE;
+		perform_close (self, NULL, NULL);
+	}
+}
+
 static void
 gcr_system_prompt_prompt_iface (GcrPromptIface *iface)
 {
@@ -1117,6 +1304,7 @@ gcr_system_prompt_prompt_iface (GcrPromptIface *iface)
 	iface->prompt_password_finish = gcr_system_prompt_password_finish;
 	iface->prompt_confirm_async = gcr_system_prompt_confirm_async;
 	iface->prompt_confirm_finish = gcr_system_prompt_confirm_finish;
+	iface->prompt_close = gcr_system_prompt_real_close;
 }
 
 /**
@@ -1303,8 +1491,8 @@ gcr_system_prompt_open_for_prompter (const gchar *prompter_name,
  * @cancellable: an optional cancellation object
  * @error: location to place an error on failure
  *
- * Close this prompt. After calling this function, no further methods may be
- * called on this object. The prompt object is not unreferenced by this
+ * Close this prompt. After calling this function, no further prompts will
+ * succeed on this object. The prompt object is not unreferenced by this
  * function, and you must unreference it once done.
  *
  * This call may block, use the gcr_system_prompt_close_async() to perform
@@ -1339,28 +1527,6 @@ gcr_system_prompt_close (GcrSystemPrompt *self,
 	return result;
 }
 
-static void
-on_prompter_stop_prompting (GObject *source,
-                            GAsyncResult *result,
-                            gpointer user_data)
-{
-	GSimpleAsyncResult *res = G_SIMPLE_ASYNC_RESULT (user_data);
-	GError *error = NULL;
-	GVariant *retval;
-
-	retval = g_dbus_connection_call_finish (G_DBUS_CONNECTION (source), result, &error);
-	if (error != NULL) {
-		_gcr_debug ("failed to stop prompting: %s", egg_error_message (error));
-		g_clear_error (&error);
-	}
-
-	if (retval)
-		g_variant_unref (retval);
-
-	g_simple_async_result_complete (res);
-	g_object_unref (res);
-}
-
 /**
  * gcr_system_prompt_close_async:
  * @self: the prompt
@@ -1382,65 +1548,20 @@ gcr_system_prompt_close_async (GcrSystemPrompt *self,
 {
 	GSimpleAsyncResult *res;
 	CallClosure *closure;
-	gboolean called = FALSE;
 
 	g_return_if_fail (GCR_SYSTEM_PROMPT (self));
 	g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));
-	g_return_if_fail (self->pv->connection != NULL);
-	g_return_if_fail (self->pv->prompt_path != NULL);
-	g_return_if_fail (self->pv->closed == FALSE);
-
-	_gcr_debug ("closing prompt");
-	self->pv->closed = TRUE;
-
-	if (self->pv->pending) {
-		res = g_object_ref (self->pv->pending);
-		g_clear_object (&self->pv->pending);
-		closure = g_simple_async_result_get_op_res_gpointer (res);
-		g_cancellable_cancel (closure->cancellable);
-		g_simple_async_result_complete_in_idle (res);
-		g_object_unref (res);
-	}
 
 	res = g_simple_async_result_new (NULL, callback, user_data,
 	                                 gcr_system_prompt_close_async);
-	closure = g_new0 (CallClosure, 1);
-	closure->cancellable = cancellable ? g_object_ref (cancellable) : g_cancellable_new ();
+	closure = call_closure_new (cancellable);
 	closure->context = g_main_context_get_thread_default ();
 	if (closure->context != NULL)
 		g_main_context_ref (closure->context);
 	g_simple_async_result_set_op_res_gpointer (res, closure, call_closure_free);
 
-	if (self->pv->prompt_registered) {
-		g_dbus_connection_unregister_object (self->pv->connection,
-		                                     self->pv->prompt_registered);
-		self->pv->prompt_registered = 0;
-	}
+	perform_close (self, res, closure->cancellable);
 
-	if (self->pv->begun_prompting) {
-		if (self->pv->connection) {
-			g_dbus_connection_call (self->pv->connection,
-			                        self->pv->prompter_bus_name,
-			                        GCR_DBUS_PROMPTER_OBJECT_PATH,
-			                        GCR_DBUS_PROMPTER_INTERFACE,
-			                        GCR_DBUS_PROMPTER_METHOD_STOP,
-			                        g_variant_new ("(o)", self->pv->prompt_path),
-			                        G_VARIANT_TYPE ("()"),
-			                        G_DBUS_CALL_FLAGS_NO_AUTO_START,
-			                        -1, closure->cancellable,
-			                        on_prompter_stop_prompting,
-			                        g_object_ref (res));
-			called = TRUE;
-		}
-	}
-
-	g_free (self->pv->prompt_path);
-	self->pv->prompt_path = NULL;
-
-	g_clear_object (&self->pv->connection);
-
-	if (!called)
-		g_simple_async_result_complete_in_idle (res);
 	g_object_unref (res);
 }
 
